# -*- coding: utf-8 -*-
"""
Created on Sat Mar 14 16:07:36 2026

@author: trent
"""

from threading import Thread

import typer
import matplotlib.pyplot as plt

app = typer.Typer(no_args_is_help=True)

from glyphloom.generation.draw import Glyph
from glyphloom.cli.commands import (fe)

app.add_typer(fe.app, name='5e')


def wait_for_input(flag):
    input("Press Enter to continue...")
    flag['done'] = True
    
    
@app.command()
def char(character_series: str = typer.Argument("The quick brown fox jumped over the lazy dog."),
        character_series_opt: str = typer.Option(None,
                                                 '-c',
                                                 '--characters',
                                                 help='Sentence or phrase you want represented, if using spaces wrap in double quotes "Hello World!"'),
        title: bool = typer.Option(True, '-t', '--hide-title', help="Flag to suppress the figure title")):
    
    done = {"done": False}
    t = Thread(target=wait_for_input, args=(done,), daemon=True)
    t.start()
    
    series = character_series or character_series_opt
    fig = Glyph.character_series(series=series,
                                 title=title)
    fig.show()
    
    while not done['done'] and plt.fignum_exists(fig.number):
        plt.pause(0.1)
        
if __name__ == "__main__":
    app()


# -*- coding: utf-8 -*-
"""
Created on Fri Mar 13 21:46:26 2026

@author: trent
"""
import typing
import re
import sys
from itertools import product, combinations, chain

import yaml

def powerset(iterable: typing.Iterable, max_r: int = 3) -> list:
    """
    Generates the powerset of sequence

    """
    sequence = list(iterable)
    result = chain.from_iterable(combinations(sequence, r)
                                 for r in range(max_r+1))
    return list(result)
    

class SpellAttribute:
    
    def __init__(self, 
                 order: int,
                 options: typing.Iterable,
                 default: typing.Any = None,
                 glyph: bool = True):
        assert not isinstance(options, str)
        
        self.order = order
        self.options = options
        if isinstance(options, dict):
            self.default = None
        else:
            self.default = options[0] if default is None else default
        self.glyph = glyph
        self.name = None
            
    def __set_name__(self, owner, name):
        self.name = name
        return
    
    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.__dict__.get(self.name, self.default)
    
    def __set__(self, instance, value):
        instance.__dict__[self.name] = value

    
    def __repr__(self) -> str:
        return (f'SpellAttribute('
                f'order={self.order}, options={self.options}, '
                f'glyph={self.glyph})')
    
    def __str__(self) -> str:
        return f'<{self.name}-SpellAttribute: default={self.default}>'
    

class SpellData:
    """
    Generic data container for n-attribute defined spells
    
    """
    name = ''
    
    def __init_subclass__(cls, **kwargs):
        attributes = cls.collect_attributes(cls)
        fields = []
        for name, value in attributes.items():
            type_hint = typing.Literal[*value.options]
            default = value.default
            fields.append((name, type_hint, default))
            
        #---- Define __init__
        params = ['name:str = ""']
        body_data = ['self.name = name']
        for name, tp, default in fields:
            search = re.search("<class '(.*)'>", str(tp))
            if search:
                tp = (search.group(1))
            if default is ...:
                params.append(f'{name}:{tp}')
            else:
                if isinstance(default, str):
                    default = repr(default)
                params.append(f'{name}:{tp} = {default}')
            body_data.append(f'self.{name} = {name}')
            
        
        param_str = ', '.join(params)
        body = '\n\t'.join(body_data)
        src = f'def __init__(self, {param_str}, **kwargs):\n\t{body}'
        module = sys.modules[cls.__module__]
        namespace = module.__dict__.copy()
        exec(src, namespace)
        setattr(cls, '__init__', namespace['__init__'])

        return
    
    @classmethod
    def collect_attributes(cls, target:type = None) -> dict[str, SpellAttribute]:
        if target is None:
            target = cls
        options = {name: getattr(target, name) 
                   for name in target.__dict__.keys()
                   if isinstance(getattr(target, name), SpellAttribute)}
        return options
    
    @classmethod
    def from_yaml(cls, yaml_filepath: str) -> type:
        with open(yaml_filepath, 'r') as fid:
            data = yaml.safe_load(fid)
            
        config = data['SpellDataConfig']
        name = config['name']
        class_name = f'SpellData_{name}'
        class_map = {'system' : name}
        for i, (attribute, attr_data) in enumerate(config['attributes'].items()):
            option_type = attr_data['option_type']
            options = attr_data['options']
            if option_type == 'range':
                options = tuple(range(*options))
            elif option_type == 'map':
                option_list = []
                for key, value in options.items():
                    if any(isinstance(i, list) for i in value):
                        # Nested
                        value = list(product(*value))
                        option_list.extend(list(map(
                            lambda t: f'{key} {str(t)}',
                            value)))
                    else:
                        option_list.extend(list(map(
                            lambda t: f'{key} ({str(t)})',
                            value)))
                options = option_list
            elif 'powerset' in option_type:
                try:
                    r = int(option_type.split('-')[-1])
                except (TypeError, IndexError):
                    r = 3
                power_set = powerset(options, r)
                if 'None' in options:
                    options = ['None', *[opt[0] if len(opt) == 1 else opt 
                                         for opt in power_set
                                         if 'None' not in opt and opt]]
                else:
                    options = [*[opt[0] if len(opt) == 1 else opt 
                                         for opt in power_set
                                         if 'None' not in opt and opt]]
                
                
            default = attr_data['default']
            glyph = attr_data.get('glyph', True)
            attr = SpellAttribute(i, options, default=default, glyph=glyph)
            setattr(attr, 'name', attribute)
            class_map[attribute] = attr
        
        for key, value in class_map.items():
            if isinstance(value, typing.Iterable) and not isinstance(value, str):
                class_map[key] = tuple(value)
        
        target = type(class_name, (SpellData,), class_map)
        
        return target
    
    @classmethod
    def yaml_spell(cls, yaml_file:str|typing.IO[str]) -> typing.Self:
        if isinstance(yaml_file, str):
            with open(yaml_file, 'r') as fid:
                data = yaml.safe_load(fid)
        else:
            data = yaml.safe_load(yaml_file)
        
        for key, value in data.items():
            if value == 'None':
                data[key] = None
        if data['system'] != cls.system:
            return None
        return cls(**data)
        
    def __repr__(self):
        arg = [f'{attr_name}={getattr(self, attr_name)}' 
               for attr_name in list(getattr(self, '__dict__', {})) 
               if not re.search('__.*__', attr_name)]
        arg_str = ', '.join(arg)
        return f'{type(self).__name__}({arg_str})'
    

def spelldata(target:type) -> SpellData:
    new_class = type(target.__name__, 
                     (SpellData, target),
                     {})
    return new_class


# -*- coding: utf-8 -*-
"""
Created on Sun Mar 15 14:23:40 2026

@author: trent
"""
import typing

import matplotlib
try:
    matplotlib.use('TkAgg')
except:
    pass
import matplotlib.pyplot as plt
plt.rcParams.update({
    "figure.facecolor":  (0.0, 0.0, 0.0, 0.0),
    "axes.facecolor":    (1.0, 1.0, 1.0, 0.0),
    "savefig.facecolor": (0.0, 0.0, 0.0, 0.0),
})
from numpy import linspace

from glyphloom.generation import data, geometry


class Glyph:
    
    def __init__(self,
                 spelldata: data.SpellData = None,
                 leylines: geometry.Leylines = None):
    
        self.spelldata = spelldata
        if leylines:
            self.leylines = leylines
        else:
            self.leylines = geometry.Leylines(
                geometry.Founts(n=len(spelldata.collect_attributes())))
        
    
    def draw(self, 
             title: bool = True, 
             title_kwargs: dict = {}, 
             legend: bool = False, 
             legend_kwargs: dict = {}, 
             ax:plt.Axes = None) -> None:
        # TODO: Add thematic colors
        if not ax:
            fig = plt.figure(constrained_layout=True)
            ax = fig.add_subplot(111)
        elif ax:
            fig = plt.gcf()
        ax.set_aspect('equal', adjustable='box')
        
        if title and not title_kwargs:
            ax.set_title(self.spelldata.name, fontdict=dict(family='Consolas'))
        elif title:
            ax.set_title(self.spelldata.name, **title_kwargs)
        
        if not self.leylines.founts.n_points:
            return fig
        
        attrs = self.spelldata.collect_attributes().values()
        cmap = plt.get_cmap('cubehelix')
        colors = cmap(linspace(0.1, 0.8, len(attrs)))
        non_glyph = 0
        for attr in attrs:
            value = getattr(self.spelldata, attr.name)
            if isinstance(value, typing.Iterable) and not isinstance(value, str):
                value = tuple(value)
            if not attr.glyph:
                if value and non_glyph:
                    kwargs = dict(fillstyle='none', 
                                  markersize=20+10*non_glyph,
                                  color='#991B1E')
                    ax.plot(0,0, 'o', **kwargs)
                elif value:
                    kwargs = dict(markersize=15, color='#991B1E')
                    ax.plot(0,0, 'o', **kwargs)
                
                non_glyph+=1
                continue
            order = attr.order
            options = attr.options
            
            if value is None: continue
            index = list(options).index(value)
            binary = self.leylines.necklace[index]
            paths = self.leylines.default_curves[order, binary.astype(bool)]
            for i, path in enumerate(paths):
                if i:
                    ax.plot(*path, color=colors[order])
                else:
                    ax.plot(*path, color=colors[order], label=f'{attr.name: <15}- {value}')
        ax.plot(*self.leylines.founts[:,1:], 'ok', fillstyle='full', markersize=6)
        ax.plot(*self.leylines.founts[:,0], 'ok', fillstyle='none', markersize=6)
        if legend and legend_kwargs:
            ax.legend(**legend_kwargs)
        elif legend:
            ax.legend(loc='lower center',
                      bbox_to_anchor=(0.5, -0.4),
                      prop={'family':'monospace'})

        ax.axis('off')
        
        return fig
    
    
    @staticmethod
    def character_series(series:str,
                         options:typing.Iterable[str] = list("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnoqprstuvwxyz.!?,;:\"'"),
                         leyline_expression:str|tuple[str] = 'non-centre-circle',
                         title:bool = True) -> None:
        figure, axes = plt.subplots(nrows=1,ncols=len(series.split()),
                                    figsize=(max(1.5*len(series.split()),5), 2))
        if title:
            figure.suptitle(series, fontdict=dict(family='Consolas'))
        if not isinstance(axes, typing.Iterable):
            axes = [axes]
        for i, word in enumerate(series.split()):
            # l_word = word.lower()
            l = len(word)
            Data = type(f"Word_{l}", (data.SpellData,), 
                        {f'letter{j}' : data.SpellAttribute(
                            j, 
                            options, 
                            default=None)
                        for j in range(l)})
            spelldata = Data(word, *word)
            
            points = l
            while True:
                try:
                    Glyph(spelldata, 
                          geometry.Leylines(
                              founts=geometry.Founts(n_points=points*2+1),
                              expression=leyline_expression
                              )
                          ).draw(title=False, ax=axes[i], legend=False)
                    break
                except IndexError:
                    points += 1
                    
        return figure
                    
if __name__ == '__main__':
    # from glyphloom.data.fifth_edition import SpellData_5e
    
    # spelldata = SpellData_5e.get_spell('Magic Stone', 
    #                                    'offline')
    # leylines = geometry.Leylines(
    #     founts=geometry.Founts(
    #         n_points=13,
    #         expression=None),
    #     expression='linear')
    # Glyph(spelldata, leylines).draw(title=False,legend=False)
    
    Glyph.character_series("The quick brown fox jumped over the lazy dog")